import logging
import re

from sqlglot import expressions as exp
from sqlglot.dialects.databricks import Databricks as SqlglotDatabricks
from sqlglot.dialects.hive import Hive
from sqlglot.dialects.dialect import if_sql
from sqlglot.dialects.dialect import rename_func
from sqlglot.errors import UnsupportedError
from sqlglot.helper import apply_index_offset, csv

from databricks.labs.lakebridge.transpiler.sqlglot import local_expression
from databricks.labs.lakebridge.transpiler.sqlglot.lca_utils import unalias_lca_in_select

# pylint: disable=too-many-public-methods

logger = logging.getLogger(__name__)

VALID_DATABRICKS_TYPES = {
    "BIGINT",
    "BINARY",
    "BOOLEAN",
    "DATE",
    "DECIMAL",
    "DOUBLE",
    "FLOAT",
    "INT",
    "INTERVAL",
    "VOID",
    "SMALLINT",
    "STRING",
    "TIMESTAMP",
    "TINYINT",
    "ARRAY",
    "MAP",
    "STRUCT",
}

PRECISION_CONST = 38
SCALE_CONST = 0


def timestamptrunc_sql(self, expression: exp.TimestampTrunc) -> str:
    return self.func("DATE_TRUNC", exp.Literal.string(expression.text("unit").upper()), self.sql(expression.this))


def _parm_sfx(self, expression: local_expression.Parameter) -> str:
    this = self.sql(expression, "this")
    this = f"{{{this}}}" if expression.args.get("wrapped") else f"{this}"
    suffix = self.sql(expression, "suffix")
    PARAMETER_TOKEN = "$"  # noqa: N806 pylint: disable=invalid-name
    return f"{PARAMETER_TOKEN}{this}{suffix}"


def _lateral_bracket_sql(self, expression: local_expression.Bracket) -> str:
    """Overwrites `sqlglot/generator.py` `bracket_sql()` function
    to convert <TABLE_ALIAS>`[COL_NAME]` to <TABLE_ALIAS>`.COL_NAME`.
    Example: c[val] ==> c.val
    """
    expressions = apply_index_offset(expression.this, expression.expressions, self.dialect.INDEX_OFFSET)
    expressions = [self.sql(e.alias_or_name.strip("'")) for e in expressions]
    # If expression contains space in between encode it in backticks(``):
    # e.g. ref."ID Number" -> ref.`ID Number`.
    expressions_sql = ", ".join(f"`{e}`" if " " in e else e for e in expressions)
    return f"{self.sql(expression, 'this')}:{expressions_sql}"


def _format_create_sql(self, expression: exp.Create) -> str:
    expression = expression.copy()

    # Remove modifiers in order to simplify the schema.  For example, this removes things like "IF NOT EXISTS"
    # from "CREATE TABLE foo IF NOT EXISTS".
    args_to_delete = ["temporary", "transient", "external", "exists", "unique", "materialized", "properties"]
    for arg_to_delete in args_to_delete:
        if expression.args.get(arg_to_delete):
            del expression.args[arg_to_delete]

    return self.create_sql(expression)


def _curr_time():
    return "date_format(current_timestamp(), 'HH:mm:ss')"


def _select_contains_index(expression: exp.Select) -> bool:
    for expr in expression.expressions:
        column = expr.unalias() if isinstance(expr, exp.Alias) else expr
        if column.name == "index":
            return True
    return False


def _has_parse_json(expression):
    if expression.find(exp.ParseJSON):
        return True
    _select = expression.find_ancestor(exp.Select)
    if _select:
        _from = _select.find(exp.From)
        if _from:
            _parse_json = _from.find(exp.ParseJSON)
            if _parse_json:
                return True
    return False


def _generate_function_str(select_contains_index, has_parse_json, generator_expr, alias, is_outer, alias_str):
    if select_contains_index:
        generator_function_str = f"POSEXPLODE({generator_expr})"
        alias_str = f"{' ' + alias.name if isinstance(alias, exp.TableAlias) else ''} AS index, value"
    elif has_parse_json and is_outer:
        generator_function_str = f"VARIANT_EXPLODE_OUTER({generator_expr})"
    elif has_parse_json:
        generator_function_str = f"VARIANT_EXPLODE({generator_expr})"
    else:
        generator_function_str = f"VIEW EXPLODE({generator_expr})"

    return generator_function_str, alias_str


def _generate_lateral_statement(self, select_contains_index, has_parse_json, generator_function_str, alias_str):
    if select_contains_index:
        lateral_statement = self.sql(f"LATERAL VIEW OUTER {generator_function_str}{alias_str}")
    elif has_parse_json:
        lateral_statement = self.sql(f", LATERAL {generator_function_str}{alias_str}")
    else:
        lateral_statement = self.sql(f" LATERAL {generator_function_str}{alias_str}")

    return lateral_statement


def _lateral_view(self: SqlglotDatabricks.Generator, expression: exp.Lateral) -> str:
    has_parse_json = _has_parse_json(expression)
    this = expression.args['this']
    alias = expression.args['alias']
    alias_str = f" AS {alias.name}" if isinstance(alias, exp.TableAlias) else ""
    generator_function_str = self.sql(this)
    is_outer = False
    select_contains_index = False

    if isinstance(this, exp.Explode):
        explode_expr = this
        parent_select = explode_expr.parent_select
        select_contains_index = _select_contains_index(parent_select) if parent_select else False
        generator_expr = ""
        if isinstance(explode_expr.this, exp.Kwarg):
            generator_expr = self.sql(explode_expr.this, 'expression')
            if not isinstance(explode_expr.this.expression, exp.ParseJSON):
                generator_expr = generator_expr.replace("{", "").replace("}", "")
        for expr in explode_expr.expressions:
            node = str(expr.this).upper()
            if node == "PATH":
                generator_expr += "." + self.sql(expr, 'expression').replace("'", "")
            if node == "OUTER":
                is_outer = True

        if not generator_expr:
            generator_expr = expression.this.this

        generator_function_str, alias_str = _generate_function_str(
            select_contains_index, has_parse_json, generator_expr, alias, is_outer, alias_str
        )

    alias_cols = alias.columns if alias else []
    if len(alias_cols) <= 2:
        alias_str = f" As {', '.join([item.this for item in alias_cols])}"

    lateral_statement = _generate_lateral_statement(
        self, select_contains_index, has_parse_json, generator_function_str, alias_str
    )
    return lateral_statement


# [TODO] Add more datatype coverage https://docs.databricks.com/sql/language-manual/sql-ref-datatypes.html
def _datatype_map(self, expression) -> str:
    if expression.this in [exp.DataType.Type.VARCHAR, exp.DataType.Type.NVARCHAR, exp.DataType.Type.CHAR]:
        return "STRING"
    if expression.this in [exp.DataType.Type.TIMESTAMP, exp.DataType.Type.TIMESTAMPLTZ]:
        return "TIMESTAMP"
    if expression.this == exp.DataType.Type.BINARY:
        return "BINARY"
    if expression.this == exp.DataType.Type.NCHAR:
        return "STRING"
    return self.datatype_sql(expression)


def try_to_date(self, expression: local_expression.TryToDate):
    func = "TRY_TO_TIMESTAMP"
    time_format = self.sql(expression, "format")
    if not time_format:
        time_format = Hive.DATE_FORMAT

    ts_result = self.func(func, expression.this, time_format)
    return exp.Date(this=ts_result)


def try_to_number(self, expression: local_expression.TryToNumber):
    func = "TRY_TO_NUMBER"
    precision = self.sql(expression, "precision")
    scale = self.sql(expression, "scale")

    if not precision:
        precision = 38

    if not scale:
        scale = 0

    func_expr = self.func(func, expression.this)
    if expression.expression:
        func_expr = self.func(func, expression.this, expression.expression)
    else:
        func_expr = expression.this

    return f"CAST({func_expr} AS DECIMAL({precision}, {scale}))"


def _to_boolean(self: SqlglotDatabricks.Generator, expression: local_expression.ToBoolean) -> str:
    this = self.sql(expression, "this")
    logger.debug(f"Converting {this} to Boolean")
    raise_error = self.sql(expression, "raise_error")
    raise_error_str = "RAISE_ERROR('Invalid parameter type for TO_BOOLEAN')" if bool(int(raise_error)) else "NULL"
    transformed = f"""
    CASE
       WHEN {this} IS NULL THEN NULL
       WHEN TYPEOF({this}) = 'boolean' THEN BOOLEAN({this})
       WHEN TYPEOF({this}) = 'string' THEN
           CASE
               WHEN LOWER({this}) IN ('true', 't', 'yes', 'y', 'on', '1') THEN TRUE
               WHEN LOWER({this}) IN ('false', 'f', 'no', 'n', 'off', '0') THEN FALSE
               ELSE RAISE_ERROR('Boolean value of x is not recognized by TO_BOOLEAN')
               END
       WHEN TRY_CAST({this} AS DOUBLE) IS NOT NULL THEN
           CASE
               WHEN ISNAN(CAST({this} AS DOUBLE)) OR CAST({this} AS DOUBLE) = DOUBLE('infinity') THEN
                    RAISE_ERROR('Invalid parameter type for TO_BOOLEAN')
               ELSE CAST({this} AS DOUBLE) != 0.0
               END
       ELSE {raise_error_str}
       END
    """
    return transformed


def _is_integer(self: SqlglotDatabricks.Generator, expression: local_expression.IsInteger) -> str:
    this = self.sql(expression, "this")
    transformed = f"""
    CASE
       WHEN {this} IS NULL THEN NULL
       WHEN {this} RLIKE '^-?[0-9]+$' AND TRY_CAST({this} AS INT) IS NOT NULL THEN TRUE
       ELSE FALSE
       END
    """
    return transformed


def _parse_json_extract_path_text(
    self: SqlglotDatabricks.Generator, expression: local_expression.JsonExtractPathText
) -> str:
    this = self.sql(expression, "this")
    path_name = expression.args["path_name"]
    if path_name.is_string:
        path = f"{self.dialect.QUOTE_START}$.{expression.text('path_name')}{self.dialect.QUOTE_END}"
    else:
        path = f"CONCAT('$.', {self.sql(expression, 'path_name')})"
    return f"GET_JSON_OBJECT({this}, {path})"


def _array_construct_compact(
    self: SqlglotDatabricks.Generator, expression: local_expression.ArrayConstructCompact
) -> str:
    exclude = "ARRAY(NULL)"
    array_expr = f"ARRAY({self.expressions(expression, flat=True)})"
    return f"ARRAY_EXCEPT({array_expr}, {exclude})"


def _array_slice(self: SqlglotDatabricks.Generator, expression: local_expression.ArraySlice) -> str:
    from_expr = self.sql(expression, "from")
    # In Databricks: array indices start at 1 in function `slice(array, start, length)`
    parsed_from_expr = 1 if from_expr == "0" else from_expr

    to_expr = self.sql(expression, "to")
    # Convert string expression to number and check if it is negative number
    if int(to_expr) < 0:
        err_message = "In Databricks: function `slice` length must be greater than or equal to 0"
        raise UnsupportedError(err_message)

    func = "SLICE"
    func_expr = self.func(func, expression.this, exp.Literal.number(parsed_from_expr), expression.args["to"])
    return func_expr


def _to_command(self, expr: exp.Command):
    this_sql = self.sql(expr, 'this')
    expression = self.sql(expr.expression, 'this')
    prefix = f"-- {this_sql}"
    if this_sql == "!":
        return f"{prefix}{expression}"
    return f"{prefix} {expression}"


def _parse_json(self, expression: exp.ParseJSON) -> str:
    return self.func("PARSE_JSON", expression.this, expression.expression)


def _to_number(self, expression: local_expression.ToNumber):
    func = "TO_NUMBER"
    precision = self.sql(expression, "precision")
    scale = self.sql(expression, "scale")

    func_expr = expression.this
    # if format is provided, else it will be vanilla cast to decimal
    if expression.expression:
        func_expr = self.func(func, expression.this, expression.expression)
        if precision:
            return f"CAST({func_expr} AS DECIMAL({precision}, {scale}))"
        return func_expr
    if not precision:
        precision = 38
    if not scale:
        scale = 0
    if not expression.expression and not precision:
        exception_msg = f"""Error Parsing expression {expression}:
                         * `format`: is required in Databricks [mandatory]
                         * `precision` and `scale`: are considered as (38, 0) if not specified.
                      """
        raise UnsupportedError(exception_msg)

    precision = PRECISION_CONST if not precision else precision
    scale = SCALE_CONST if not scale else scale
    return f"CAST({func_expr} AS DECIMAL({precision}, {scale}))"


def _uuid(self: SqlglotDatabricks.Generator, expression: local_expression.UUID) -> str:
    namespace = self.sql(expression, "this")
    name = self.sql(expression, "name")

    if namespace and name:
        logger.warning("UUID version 5 is not supported currently. Needs manual intervention.")
        return f"UUID({namespace}, {name})"

    return "UUID()"


def _parse_date_trunc(self: SqlglotDatabricks.Generator, expression: local_expression.DateTrunc) -> str:
    if not expression.args.get("unit"):
        error_message = f"Required keyword: 'unit' missing for {exp.DateTrunc}"
        raise UnsupportedError(error_message)
    return self.func("TRUNC", expression.this, expression.args.get("unit"))


def _get_within_group_params(
    expr: exp.ArrayAgg | exp.GroupConcat,
    within_group: exp.WithinGroup,
) -> local_expression.WithinGroupParams:
    has_distinct = isinstance(expr.this, exp.Distinct)
    agg_col = expr.this.expressions[0] if has_distinct else expr.this
    order_clause = within_group.expression
    order_cols = []
    for e in order_clause.expressions:
        desc = e.args.get("desc")
        is_order_a = not desc or exp.false() == desc
        order_cols.append((e.this, is_order_a))
    return local_expression.WithinGroupParams(
        agg_col=agg_col,
        order_cols=order_cols,
    )


def _create_named_struct_for_cmp(wg_params: local_expression.WithinGroupParams) -> exp.Expression:
    agg_col = wg_params.agg_col
    order_kv = []
    for i, (col, _) in enumerate(wg_params.order_cols):
        order_kv.extend([exp.Literal(this=f"sort_by_{i}", is_string=True), col])

    named_struct_func = exp.Anonymous(
        this="named_struct",
        expressions=[
            exp.Literal(this="value", is_string=True),
            agg_col,
            *order_kv,
        ],
    )
    return named_struct_func


def _current_date(self, expression: exp.CurrentDate) -> str:
    zone = self.sql(expression, "this")
    return f"CURRENT_DATE({zone})" if zone else "CURRENT_DATE()"


def _not_sql(self, expression: exp.Not) -> str:
    if isinstance(expression.this, exp.Is):
        return f"{self.sql(expression.this, 'this')} IS NOT {self.sql(expression.this, 'expression')}"
    return f"NOT {self.sql(expression, 'this')}"


def to_array(self, expression: exp.ToArray) -> str:
    return f"IF({self.sql(expression.this)} IS NULL, NULL, {self.func('ARRAY', expression.this)})"


class Databricks(SqlglotDatabricks):  #
    # Instantiate Databricks Dialect
    databricks = SqlglotDatabricks()
    NULL_ORDERING = "nulls_are_small"

    class Generator(SqlglotDatabricks.Generator):
        INVERSE_TIME_MAPPING: dict[str, str] = {
            **{v: k for k, v in SqlglotDatabricks.TIME_MAPPING.items()},
            "%-d": "dd",
        }

        COLLATE_IS_FUNC = True
        # [TODO]: Variant needs to be transformed better, for now parsing to string was deemed as the choice.
        TYPE_MAPPING = {
            **SqlglotDatabricks.Generator.TYPE_MAPPING,
            exp.DataType.Type.TINYINT: "TINYINT",
            exp.DataType.Type.SMALLINT: "SMALLINT",
            exp.DataType.Type.INT: "INT",
            exp.DataType.Type.BIGINT: "BIGINT",
            exp.DataType.Type.DATETIME: "TIMESTAMP",
            exp.DataType.Type.VARCHAR: "STRING",
            exp.DataType.Type.VARIANT: "VARIANT",
            exp.DataType.Type.FLOAT: "DOUBLE",
            exp.DataType.Type.OBJECT: "STRING",
            exp.DataType.Type.GEOGRAPHY: "STRING",
        }

        TRANSFORMS = {
            **SqlglotDatabricks.Generator.TRANSFORMS,
            exp.Create: _format_create_sql,
            exp.DataType: _datatype_map,
            exp.CurrentTime: _curr_time(),
            exp.Lateral: _lateral_view,
            exp.FromBase64: rename_func("UNBASE64"),
            exp.AutoIncrementColumnConstraint: lambda *_: "GENERATED ALWAYS AS IDENTITY",
            local_expression.Parameter: _parm_sfx,
            local_expression.ToBoolean: _to_boolean,
            local_expression.Bracket: _lateral_bracket_sql,
            local_expression.MakeDate: rename_func("MAKE_DATE"),
            local_expression.TryToDate: try_to_date,
            local_expression.TryToNumber: try_to_number,
            local_expression.IsInteger: _is_integer,
            local_expression.JsonExtractPathText: _parse_json_extract_path_text,
            local_expression.BitOr: rename_func("BIT_OR"),
            local_expression.ArrayConstructCompact: _array_construct_compact,
            local_expression.ArrayIntersection: rename_func("ARRAY_INTERSECT"),
            local_expression.ArraySlice: _array_slice,
            local_expression.ObjectKeys: rename_func("JSON_OBJECT_KEYS"),
            exp.ParseJSON: _parse_json,
            local_expression.TimestampFromParts: rename_func("MAKE_TIMESTAMP"),
            local_expression.ToDouble: rename_func("DOUBLE"),
            exp.Rand: rename_func("RANDOM"),
            local_expression.ToVariant: rename_func("TO_JSON"),
            local_expression.ToObject: rename_func("TO_JSON"),
            exp.ToBase64: rename_func("BASE64"),
            local_expression.ToNumber: _to_number,
            local_expression.UUID: _uuid,
            local_expression.DateTrunc: _parse_date_trunc,
            exp.ApproxQuantile: rename_func("APPROX_PERCENTILE"),
            exp.TimestampTrunc: timestamptrunc_sql,
            exp.Mod: rename_func("MOD"),
            exp.NullSafeEQ: lambda self, e: self.binary(e, "<=>"),
            exp.If: if_sql(false_value="NULL"),
            exp.Command: _to_command,
            exp.CurrentDate: _current_date,
            exp.Not: _not_sql,
            local_expression.ToArray: to_array,
            local_expression.ArrayExists: rename_func("EXISTS"),
        }

        def preprocess(self, expression: exp.Expression) -> exp.Expression:
            fixed_ast = expression.transform(unalias_lca_in_select, copy=False)
            return super().preprocess(fixed_ast)

        def format_time(self, expression: exp.Expression, inverse_time_mapping=None, inverse_time_trie=None):
            return super().format_time(expression, self.INVERSE_TIME_MAPPING)

        def join_sql(self, expression: exp.Join) -> str:
            """Overwrites `join_sql()` in `sqlglot/generator.py`
            Added logic to handle Lateral View
            """
            op_list = [
                expression.method,
                "GLOBAL" if expression.args.get("global") else None,
                expression.side,
                expression.kind,
                expression.hint if self.JOIN_HINTS else None,
            ]

            op_sql = " ".join(op for op in op_list if op)
            on_sql = self.sql(expression, "on")
            using = expression.args.get("using")

            if not on_sql and using:
                on_sql = csv(*(self.sql(column) for column in using))

            this_sql = self.sql(expression, "this")

            if on_sql:
                on_sql = self.indent(on_sql, skip_first=True)
                space = self.seg(" " * self.pad) if self.pretty else " "
                if using:
                    on_sql = f"{space}USING ({on_sql})"
                else:
                    on_sql = f"{space}ON {on_sql}"
            # Added the below elif block to handle Lateral View clause
            elif not op_sql and isinstance(expression.this, exp.Lateral):
                return f"\n {this_sql}"
            elif not op_sql:
                return f", {this_sql}"

            op_sql = f"{op_sql} JOIN" if op_sql else "JOIN"
            return f"{self.seg(op_sql)} {this_sql}{on_sql}"

        def arrayagg_sql(self, expression: exp.ArrayAgg) -> str:
            sql = self.func("ARRAY_AGG", expression.this)
            within_group = expression.parent if isinstance(expression.parent, exp.WithinGroup) else None
            if not within_group:
                return sql

            wg_params = _get_within_group_params(expression, within_group)
            if len(wg_params.order_cols) == 1:
                order_col, is_order_asc = wg_params.order_cols[0]
                if wg_params.agg_col == order_col:
                    return f"SORT_ARRAY({sql}{'' if is_order_asc else ', FALSE'})"

            named_struct_func = _create_named_struct_for_cmp(wg_params)
            comparisons = []
            for i, (_, is_order_asc) in enumerate(wg_params.order_cols):
                comparisons.append(
                    f"WHEN left.sort_by_{i} < right.sort_by_{i} THEN {'-1' if is_order_asc else '1'} "
                    f"WHEN left.sort_by_{i} > right.sort_by_{i} THEN {'1' if is_order_asc else '-1'}"
                )

            array_sort = self.func(
                "ARRAY_SORT",
                self.func("ARRAY_AGG", named_struct_func),
                f"""(left, right) -> CASE
                        {' '.join(comparisons)}
                        ELSE 0
                    END""",
            )
            return self.func("TRANSFORM", array_sort, "s -> s.value")

        def groupconcat_sql(self, expr: exp.GroupConcat) -> str:
            arr_agg = exp.ArrayAgg(this=expr.this)
            within_group = expr.parent.copy() if isinstance(expr.parent, exp.WithinGroup) else None
            if within_group:
                arr_agg.parent = within_group

            return self.func(
                "ARRAY_JOIN",
                arr_agg,
                expr.args.get("separator") or exp.Literal(this="", is_string=True),
            )

        def withingroup_sql(self, expression: exp.WithinGroup) -> str:
            agg_expr = expression.this
            if isinstance(agg_expr, (exp.ArrayAgg, exp.GroupConcat)):
                return self.sql(agg_expr)

            return super().withingroup_sql(expression)

        def split_sql(self, expression: local_expression.Split) -> str:
            """
            :param expression: local_expression.Split expression to be parsed
            :return: Converted expression (SPLIT) compatible with Databricks
            """
            delimiter = " "
            # To handle default delimiter
            if expression.expression:
                delimiter = expression.expression.name

            # Parsing logic to handle String and Table columns
            if expression.name and isinstance(expression.name, str):
                expr_name = f"'{expression.name}'"
            else:
                expr_name = expression.args["this"]
            return f"""SPLIT({expr_name},'[{delimiter}]')"""

        def delete_sql(self, expression: exp.Delete) -> str:
            this = self.sql(expression, "this")
            using = self.sql(expression, "using")
            where = self.sql(expression, "where")
            returning = self.sql(expression, "returning")
            limit = self.sql(expression, "limit")
            tables = self.expressions(expression, key="tables")
            tables = f" {tables}" if tables else ""

            if using:
                using = f" USING {using}" if using else ""
                where = where.replace("WHERE", "ON")
            else:
                this = f"FROM {this}" if this else ""

            if self.RETURNING_END:
                expression_sql = f" {this}{using}{where}{returning}{limit}"
            else:
                expression_sql = f"{returning}{this}{where}{limit}"

            if using:
                return self.prepend_ctes(expression, f"MERGE INTO {tables}{expression_sql} WHEN MATCHED THEN DELETE;")

            return self.prepend_ctes(expression, f"DELETE{tables}{expression_sql};")

        def converttimezone_sql(self, expression: exp.ConvertTimezone):
            func = "CONVERT_TIMEZONE"
            expr = expression.args["tgtTZ"]
            if len(expression.args) == 3 and expression.args.get("this"):
                expr = expression.args["this"]

            result = self.func(func, expression.args["srcTZ"], expr)
            if len(expression.args) == 3:
                result = self.func(func, expression.args["srcTZ"], expression.args["tgtTZ"], expr)

            return result

        def strtok_sql(self, expression: local_expression.StrTok) -> str:
            """
            :param expression: local_expression.StrTok expression to be parsed
            :return: Converted expression (SPLIT_PART) compatible with Databricks
            """
            # To handle default delimiter
            if expression.expression:
                delimiter = expression.expression.name
            else:
                delimiter = " "

            # Handle String and Table columns
            if expression.name and isinstance(expression.name, str):
                expr_name = f"'{expression.name}'"
            else:
                expr_name = expression.args["this"]

            # Handle Partition Number
            if len(expression.args) == 3 and expression.args.get("partNum"):
                part_num = expression.args["partNum"]
            else:
                part_num = 1

            return f"SPLIT_PART({expr_name}, '{delimiter}', {part_num})"

        def splitpart_sql(self, expression: local_expression.SplitPart) -> str:
            """
            :param expression: local_expression.SplitPart expression to be parsed
            :return: Converted expression (SPLIT_PART) compatible with Databricks
            """
            expr_name = self.sql(expression.this)
            delimiter = self.sql(expression.expression)
            part_num = self.sql(expression.args["partNum"])
            return f"SPLIT_PART({expr_name}, {delimiter}, {part_num})"

        def transaction_sql(self, expression: exp.Transaction) -> str:
            """
            Skip begin command
            :param expression:
            :return: Empty string for unsupported operation
            """
            return ""

        def rollback_sql(self, expression: exp.Rollback) -> str:
            """
            Skip rollback command
            :param expression:
            :return: Empty string for unsupported operation
            """
            return ""

        def commit_sql(self, expression: exp.Commit) -> str:
            """
            Skip commit command
            :param expression:
            :return: Empty string for unsupported operation
            """
            return ""

        def command_sql(self, expression: exp.Command) -> str:
            """
            Skip any session, stream, task related commands
            :param expression:
            :return: Empty string for unsupported operations or objects
            """
            filtered_commands = [
                "CREATE",
                "ALTER",
                "DESCRIBE",
                "DROP",
                "SHOW",
                "EXECUTE",
            ]
            ignored_objects = [
                "STREAM",
                "TASK",
                "STREAMS",
                "TASKS",
                "SESSION",
            ]

            command = self.sql(expression, "this").upper()
            expr = expression.text("expression").strip()
            obj = re.split(r"\s+", expr, maxsplit=2)[0].upper() if expr else ""
            if command in filtered_commands and obj in ignored_objects:
                return ""
            return f"{command} {expr}"

        def currenttimestamp_sql(self, _: exp.CurrentTimestamp) -> str:
            return self.func("CURRENT_TIMESTAMP")

        def update_sql(self, expression: exp.Update) -> str:
            this = self.sql(expression, "this")
            set_sql = self.expressions(expression, flat=True)
            from_sql = self.sql(expression, "from")
            where_sql = self.sql(expression, "where")
            returning = self.sql(expression, "returning")
            order = self.sql(expression, "order")
            limit = self.sql(expression, "limit")

            if from_sql:
                from_sql = from_sql.replace("FROM", "USING", 1)
                where_sql = where_sql.replace("WHERE", "ON")

            if self.RETURNING_END:
                expression_sql = f"{from_sql}{where_sql}{returning}"
            else:
                expression_sql = f"{returning}{from_sql}{where_sql}"

            if from_sql:
                sql = f"MERGE INTO {this}{expression_sql} WHEN MATCHED THEN UPDATE SET {set_sql}{order}{limit}"
            else:
                sql = f"UPDATE {this} SET {set_sql}{expression_sql}{order}{limit}"

            return self.prepend_ctes(expression, sql)

        def struct_sql(self, expression: exp.Struct) -> str:
            expression.set(
                "expressions",
                [
                    (
                        exp.alias_(
                            e.expression, e.name if hasattr(e.this, "is_string") and e.this.is_string else e.this
                        )
                        if isinstance(e, exp.PropertyEQ)
                        else e
                    )
                    for e in expression.expressions
                ],
            )

            return self.function_fallback_sql(expression)

        def anonymous_sql(self: SqlglotDatabricks.Generator, expression: exp.Anonymous) -> str:
            if expression.this == "EDITDISTANCE":
                return self.func("LEVENSHTEIN", *expression.expressions)
            if expression.this == "TO_TIMESTAMP":
                return self.sql(
                    exp.Cast(this=expression.expressions[0], to=exp.DataType(this=exp.DataType.Type.TIMESTAMP))
                )

            return self.func(self.sql(expression, "this"), *expression.expressions)

        def order_sql(self, expression: exp.Order, flat: bool = False) -> str:
            if isinstance(expression.parent, exp.Window):
                for ordered_expression in expression.expressions:
                    if isinstance(ordered_expression, exp.Ordered) and ordered_expression.args.get('desc') is None:
                        ordered_expression.args['desc'] = False
            return super().order_sql(expression, flat)

        def add_column_sql(self, expression: exp.Alter) -> str:
            # Final output contains ADD COLUMN before each column
            # This function will handle this issue and return the final output
            columns = self.expressions(expression, key="actions", flat=True)
            return f"ADD COLUMN {columns}"



import logging
from functools import reduce
from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.functions import col, expr, lit

from databricks.labs.lakebridge.reconcile.connectors.dialect_utils import DialectUtils
from databricks.labs.lakebridge.reconcile.exception import ColumnMismatchException
from databricks.labs.lakebridge.reconcile.recon_capture import (
    ReconIntermediatePersist,
)
from databricks.labs.lakebridge.reconcile.recon_output_config import (
    DataReconcileOutput,
    MismatchOutput,
)
from databricks.labs.lakebridge.reconcile.recon_config import (
    AggregateRule,
    ColumnMapping,
)

logger = logging.getLogger(__name__)

_HASH_COLUMN_NAME = "hash_value_recon"
_SAMPLE_ROWS = 50


def _raise_column_mismatch_exception(msg: str, source_missing: list[str], target_missing: list[str]) -> Exception:
    error_msg = (
        f"{msg}\n"
        f"columns missing in source: {','.join(source_missing) if source_missing else None}\n"
        f"columns missing in target: {','.join(target_missing) if target_missing else None}\n"
    )
    return ColumnMismatchException(error_msg)


def _generate_join_condition(source_alias, target_alias, key_columns):
    conditions = [
        col(f"{source_alias}.{DialectUtils.ansi_normalize_identifier(key_column)}").eqNullSafe(
            col(f"{target_alias}.{DialectUtils.ansi_normalize_identifier(key_column)}")
        )
        for key_column in key_columns
    ]
    return reduce(lambda a, b: a & b, conditions)


def _build_column_selector(table_name, column_name):
    alias = DialectUtils.ansi_normalize_identifier(f"{table_name}_{DialectUtils.unnormalize_identifier(column_name)}")
    return f'{table_name}.{DialectUtils.ansi_normalize_identifier(column_name)} as {alias}'


def _build_mismatch_column(table, column):
    return col(DialectUtils.ansi_normalize_identifier(column)).alias(
        DialectUtils.unnormalize_identifier(column.replace(f'{table}_', '').lower())
    )


def reconcile_data(
    source: DataFrame,
    target: DataFrame,
    key_columns: list[str],
    report_type: str,
    spark: SparkSession,
    path: str,
) -> DataReconcileOutput:
    source_alias = "src"
    target_alias = "tgt"
    if report_type not in {"data", "all"}:
        key_columns = [_HASH_COLUMN_NAME]
    df = (
        source.alias(source_alias)
        .join(
            other=target.alias(target_alias),
            on=_generate_join_condition(source_alias, target_alias, key_columns),
            how="full",
        )
        .selectExpr(
            *[f'{_build_column_selector(source_alias, col_name)}' for col_name in source.columns],
            *[f'{_build_column_selector(target_alias, col_name)}' for col_name in target.columns],
        )
    )

    # Write unmatched df to volume
    df = ReconIntermediatePersist(spark, path).write_and_read_unmatched_df_with_volumes(df)
    logger.warning(f"Unmatched data was written to {path} successfully")

    mismatch = _get_mismatch_data(df, source_alias, target_alias) if report_type in {"all", "data"} else None

    missing_in_src = (
        df.filter(col(f"{source_alias}_{_HASH_COLUMN_NAME}").isNull())
        .select(
            *[
                _build_mismatch_column(target_alias, col_name)
                for col_name in df.columns
                if col_name.startswith(f'{target_alias}_')
            ]
        )
        .drop(f"{_HASH_COLUMN_NAME}")
    )

    missing_in_tgt = (
        df.filter(col(f"{target_alias}_{_HASH_COLUMN_NAME}").isNull())
        .select(
            *[
                _build_mismatch_column(source_alias, col_name)
                for col_name in df.columns
                if col_name.startswith(f'{source_alias}_')
            ]
        )
        .drop(f"{_HASH_COLUMN_NAME}")
    )
    mismatch_count = 0
    if mismatch:
        mismatch_count = mismatch.count()

    missing_in_src_count = missing_in_src.count()
    missing_in_tgt_count = missing_in_tgt.count()

    return DataReconcileOutput(
        mismatch_count=mismatch_count,
        missing_in_src_count=missing_in_src_count,
        missing_in_tgt_count=missing_in_tgt_count,
        missing_in_src=missing_in_src.limit(_SAMPLE_ROWS),
        missing_in_tgt=missing_in_tgt.limit(_SAMPLE_ROWS),
        mismatch=MismatchOutput(mismatch_df=mismatch),
    )


def _get_mismatch_data(df: DataFrame, src_alias: str, tgt_alias: str) -> DataFrame:
    return (
        df.filter(
            (col(f"{src_alias}_{_HASH_COLUMN_NAME}").isNotNull())
            & (col(f"{tgt_alias}_{_HASH_COLUMN_NAME}").isNotNull())
        )
        .withColumn(
            "hash_match",
            col(f"{src_alias}_{_HASH_COLUMN_NAME}") == col(f"{tgt_alias}_{_HASH_COLUMN_NAME}"),
        )
        .filter(col("hash_match") == lit(False))
        .select(
            *[
                _build_mismatch_column(src_alias, col_name)
                for col_name in df.columns
                if col_name.startswith(f'{src_alias}_')
            ]
        )
        .drop(f"{_HASH_COLUMN_NAME}")
    )


def _build_capture_df(df: DataFrame) -> DataFrame:
    columns = [
        col(DialectUtils.ansi_normalize_identifier(column)).alias(DialectUtils.unnormalize_identifier(column))
        for column in df.columns
    ]
    return df.select(*columns)


def capture_mismatch_data_and_columns(source: DataFrame, target: DataFrame, key_columns: list[str]) -> MismatchOutput:
    source_df = _build_capture_df(source)
    target_df = _build_capture_df(target)
    unnormalized_key_columns = [DialectUtils.unnormalize_identifier(column) for column in key_columns]

    source_columns = source_df.columns
    target_columns = target_df.columns

    if source_columns != target_columns:
        message = "source and target should have same columns for capturing the mismatch data"
        source_missing = [column for column in target_columns if column not in source_columns]
        target_missing = [column for column in source_columns if column not in target_columns]
        raise _raise_column_mismatch_exception(message, source_missing, target_missing)

    check_columns = [column for column in source_columns if column not in unnormalized_key_columns]
    mismatch_df = _get_mismatch_df(source_df, target_df, unnormalized_key_columns, check_columns)
    mismatch_columns = _get_mismatch_columns(mismatch_df, check_columns)
    return MismatchOutput(mismatch_df, mismatch_columns)


def _get_mismatch_columns(df: DataFrame, columns: list[str]):
    # Collect the DataFrame to a local variable
    local_df = df.collect()
    mismatch_columns = []
    for column in columns:
        # Check if any row has False in the column
        if any(not row[column + "_match"] for row in local_df):
            mismatch_columns.append(column)
    return mismatch_columns


def _normalize_mismatch_df_col(column, suffix):
    unnormalized = DialectUtils.unnormalize_identifier(column) + suffix
    return DialectUtils.ansi_normalize_identifier(unnormalized)


def _unnormalize_mismatch_df_col(column, suffix):
    unnormalized = DialectUtils.unnormalize_identifier(column) + suffix
    return unnormalized


def _get_mismatch_df(source: DataFrame, target: DataFrame, key_columns: list[str], column_list: list[str]):
    source_aliased = [
        col('base.' + DialectUtils.ansi_normalize_identifier(column)).alias(
            _unnormalize_mismatch_df_col(column, '_base')
        )
        for column in column_list
    ]
    target_aliased = [
        col('compare.' + DialectUtils.ansi_normalize_identifier(column)).alias(
            _unnormalize_mismatch_df_col(column, '_compare')
        )
        for column in column_list
    ]

    match_expr = [
        expr(f"{_normalize_mismatch_df_col(column,'_base')}=={_normalize_mismatch_df_col(column,'_compare')}").alias(
            _unnormalize_mismatch_df_col(column, '_match')
        )
        for column in column_list
    ]
    key_cols = [col(DialectUtils.ansi_normalize_identifier(column)) for column in key_columns]
    select_expr = key_cols + source_aliased + target_aliased + match_expr

    logger.info(f"KEY COLUMNS: {key_columns}")
    logger.info(f"SELECT COLUMNS: {select_expr}")

    mismatch_df = (
        source.alias('base').join(other=target.alias('compare'), on=key_columns, how="inner").select(*select_expr)
    )

    compare_columns = [
        DialectUtils.ansi_normalize_identifier(column) for column in mismatch_df.columns if column not in key_columns
    ]
    return mismatch_df.select(*key_cols + sorted(compare_columns))


def _generate_agg_join_condition(source_alias: str, target_alias: str, key_columns: list[str]):
    join_columns: list[ColumnMapping] = [
        ColumnMapping(source_name=f"source_group_by_{key_col}", target_name=f"target_group_by_{key_col}")
        for key_col in key_columns
    ]
    conditions = [
        col(f"{source_alias}.{mapping.source_name}").eqNullSafe(col(f"{target_alias}.{mapping.target_name}"))
        for mapping in join_columns
    ]
    return reduce(lambda a, b: a & b, conditions)


def _agg_conditions(
    cols: list[ColumnMapping] | None,
    condition_type: str = "group_filter",
    op_type: str = "and",
):
    """
    Generate conditions for aggregated data comparison based on the condition type
    and reduces it based on the operator (and, or)

    e.g.,  cols = [(source_min_col1, target_min_col1)]
              1. condition_type = "group_filter"
                    source_group_by_col1 is not null and target_group_by_col1 is not null
              2. condition_type = "select"
                    source_min_col1 == target_min_col1
              3. condition_type = "missing_in_src"
                    source_min_col1 is null
              4. condition_type = "missing_in_tgt"
                      target_min_col1 is null

    :param cols:  List of columns to compare
    :param condition_type:  Type of condition to generate
    :param op_type: and, or
    :return:  Reduced column expressions
    """
    assert cols, "Columns must be specified for aggregation conditions"

    if condition_type == "group_filter":
        conditions_list = [
            (col(f"{mapping.source_name}").isNotNull() & col(f"{mapping.target_name}").isNotNull()) for mapping in cols
        ]
    elif condition_type == "select":
        conditions_list = [col(f"{mapping.source_name}") == col(f"{mapping.target_name}") for mapping in cols]
    elif condition_type == "missing_in_src":
        conditions_list = [col(f"{mapping.source_name}").isNull() for mapping in cols]
    elif condition_type == "missing_in_tgt":
        conditions_list = [col(f"{mapping.target_name}").isNull() for mapping in cols]
    else:
        raise ValueError(f"Invalid condition type: {condition_type}")

    return reduce(lambda a, b: a & b if op_type == "and" else a | b, conditions_list)


def _generate_match_columns(select_cols: list[ColumnMapping]):
    """
    Generate match columns for the given select columns
    e.g.,  select_cols = [(source_min_col1, target_min_col1), (source_count_col3, target_count_col3)]
            |--------------------------------------|---------------------|
           |               match_min_col1                      |  match_count_col3 |
           |--------------------------------------|--------------------|
             source_min_col1 == target_min_col1 | source_count_col3 == target_count_col3
           --------------------------------------|---------------------|

    :param select_cols:
    :return:
    """
    items = []
    for mapping in select_cols:
        match_col_name = mapping.source_name.replace("source_", "match_")
        items.append((match_col_name, col(f"{mapping.source_name}") == col(f"{mapping.target_name}")))
    return items


def _get_mismatch_agg_data(
    df: DataFrame,
    select_cols: list[ColumnMapping],
    group_cols: list[ColumnMapping] | None,
) -> DataFrame:
    # TODO:  Integrate with _get_mismatch_data function
    """
    For each rule select columns, generate a match column to compare the aggregated data between Source and Target

      e.g., select_cols = [(source_min_col1, target_min_col1), (source_count_col3, target_count_col3)]

            source_min_col1 | target_min_col1 | match_min_col1 |  agg_data_match |
            -----------------|--------------------|----------------|-------------------|
                   11    |   12    |source_min_col1 == target_min_col1 | False                     |

    :param df: Joined DataFrame with aggregated data from Source and Target
    :param select_cols:  Rule specific select columns
    :param group_cols: Rule specific group by columns, if any
    :return: DataFrame with match_<AGG_TYPE>_<COLUMN> and agg_data_match columns
                 to identify the aggregate data mismatch between Source and Target
    """
    df_with_match_cols = df

    if group_cols:
        # Filter Conditions are in the format of: source_group_by_col1 is not null and target_group_by_col1 is not null
        filter_conditions = _agg_conditions(group_cols)
        df_with_match_cols = df_with_match_cols.filter(filter_conditions)

    # Generate match columns for the select columns. e.g., match_<AGG_TYPE>_<COLUMN>
    for match_column_name, match_column in _generate_match_columns(select_cols):
        df_with_match_cols = df_with_match_cols.withColumn(match_column_name, match_column)

    # e.g., source_min_col1 == target_min_col1 and source_count_col3 == target_count_col3
    select_conditions = _agg_conditions(select_cols, "select")

    return df_with_match_cols.withColumn("agg_data_match", select_conditions).filter(
        col("agg_data_match") == lit(False)
    )


def reconcile_agg_data_per_rule(
    joined_df: DataFrame,
    source_columns: list[str],
    target_columns: list[str],
    rule: AggregateRule,
) -> DataReconcileOutput:
    """ "
    Generates the reconciliation output for the given rule
    """
    # Generates select columns in the format of:
    # [(source_min_col1, target_min_col1), (source_count_col3, target_count_col3) ... ]

    rule_select_columns = [
        ColumnMapping(
            source_name=f"source_{rule.agg_type}_{rule.agg_column}",
            target_name=f"target_{rule.agg_type}_{rule.agg_column}",
        )
    ]

    rule_group_columns = None
    if rule.group_by_columns:
        rule_group_columns = [
            ColumnMapping(source_name=f"source_group_by_{group_col}", target_name=f"target_group_by_{group_col}")
            for group_col in rule.group_by_columns
        ]
        rule_select_columns.extend(rule_group_columns)

    df_rule_columns = []
    for mapping in rule_select_columns:
        df_rule_columns.extend([mapping.source_name, mapping.target_name])

    joined_df_with_rule_cols = joined_df.selectExpr(*df_rule_columns)

    # Data mismatch between Source and Target aggregated data
    mismatch = _get_mismatch_agg_data(joined_df_with_rule_cols, rule_select_columns, rule_group_columns)

    # Data missing in Source DataFrame
    rule_target_columns = set(target_columns).intersection([mapping.target_name for mapping in rule_select_columns])

    missing_in_src = joined_df_with_rule_cols.filter(_agg_conditions(rule_select_columns, "missing_in_src")).select(
        *rule_target_columns
    )

    # Data missing in Target DataFrame
    rule_source_columns = set(source_columns).intersection([mapping.source_name for mapping in rule_select_columns])
    missing_in_tgt = joined_df_with_rule_cols.filter(_agg_conditions(rule_select_columns, "missing_in_tgt")).select(
        *rule_source_columns
    )

    mismatch_count = 0
    if mismatch:
        mismatch_count = mismatch.count()

    rule_reconcile_output = DataReconcileOutput(
        mismatch_count=mismatch_count,
        missing_in_src_count=missing_in_src.count(),
        missing_in_tgt_count=missing_in_tgt.count(),
        missing_in_src=missing_in_src.limit(_SAMPLE_ROWS),
        missing_in_tgt=missing_in_tgt.limit(_SAMPLE_ROWS),
        mismatch=MismatchOutput(mismatch_df=mismatch),
    )

    return rule_reconcile_output


def join_aggregate_data(
    source: DataFrame,
    target: DataFrame,
    key_columns: list[str] | None,
    spark: SparkSession,
    path: str,
) -> DataFrame:
    # TODO:  Integrate with reconcile_data function

    source_alias = "src"
    target_alias = "tgt"

    # Generates group by columns in the format of:
    # [(source_group_by_col1, target_group_by_col1), (source_group_by_col2, target_group_by_col2) ... ]

    if key_columns:
        # If there are Group By columns, do Full join on the grouped columns
        df = source.alias(source_alias).join(
            other=target.alias(target_alias),
            on=_generate_agg_join_condition(source_alias, target_alias, key_columns),
            how="full",
        )
    else:
        # If there is no Group By condition, do Cross join as there is only one record
        df = source.alias(source_alias).join(
            other=target.alias(target_alias),
            how="cross",
        )

    joined_df = df.selectExpr(
        *source.columns,
        *target.columns,
    )

    # Write the joined df to volume path
    joined_volume_df = ReconIntermediatePersist(spark, path).write_and_read_unmatched_df_with_volumes(joined_df).cache()
    logger.warning(f"Unmatched data is written to {path} successfully")

    return joined_volume_df



import dataclasses
import logging
from datetime import datetime, timezone, timedelta
from typing import Any

from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.installer import InstallState
from databricks.labs.blueprint.wheels import ProductInfo
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import InvalidParameterValue
from databricks.sdk.service import compute
from databricks.sdk.service.jobs import Task, PythonWheelTask, JobCluster, JobSettings, JobParameterDefinition

from databricks.labs.lakebridge.config import ReconcileConfig
from databricks.labs.lakebridge.reconcile.constants import ReconSourceType

logger = logging.getLogger(__name__)

_TEST_JOBS_PURGE_TIMEOUT = timedelta(hours=1, minutes=15)


class JobDeployment:
    def __init__(
        self,
        ws: WorkspaceClient,
        installation: Installation,
        install_state: InstallState,
        product_info: ProductInfo,
    ):
        self._ws = ws
        self._installation = installation
        self._install_state = install_state
        self._product_info = product_info

    def deploy_recon_job(self, name, recon_config: ReconcileConfig, remorph_wheel_path: str):
        logger.info("Deploying reconciliation job.")
        job_id = self._update_or_create_recon_job(name, recon_config, remorph_wheel_path)
        logger.info(f"Reconciliation job deployed with job_id={job_id}")
        logger.info(f"Job URL: {self._ws.config.host}#job/{job_id}")
        self._install_state.save()

    def _update_or_create_recon_job(self, name, recon_config: ReconcileConfig, remorph_wheel_path: str) -> str:
        description = "Run the reconciliation process"
        task_key = "run_reconciliation"

        job_settings = self._recon_job_settings(name, task_key, description, recon_config, remorph_wheel_path)
        if name in self._install_state.jobs:
            try:
                job_id = int(self._install_state.jobs[name])
                logger.info(f"Updating configuration for job `{name}`, job_id={job_id}")
                self._ws.jobs.reset(job_id, JobSettings(**job_settings))
                return str(job_id)
            except InvalidParameterValue:
                del self._install_state.jobs[name]
                logger.warning(f"Job `{name}` does not exist anymore for some reason")
                return self._update_or_create_recon_job(name, recon_config, remorph_wheel_path)

        logger.info(f"Creating new job configuration for job `{name}`")
        new_job = self._ws.jobs.create(**job_settings)
        assert new_job.job_id is not None
        self._install_state.jobs[name] = str(new_job.job_id)
        return str(new_job.job_id)

    def _recon_job_settings(
        self,
        job_name: str,
        task_key: str,
        description: str,
        recon_config: ReconcileConfig,
        remorph_wheel_path: str,
    ) -> dict[str, Any]:
        latest_lts_spark = self._ws.clusters.select_spark_version(latest=True, long_term_support=True)
        version = self._product_info.version()
        version = version if not self._ws.config.is_gcp else version.replace("+", "-")
        tags = {"version": f"v{version}"}
        if self._is_testing():
            # Add RemoveAfter tag for test job cleanup
            date_to_remove = self._get_test_purge_time()
            tags.update({"RemoveAfter": date_to_remove})

        return {
            "name": self._name_with_prefix(job_name),
            "tags": tags,
            "job_clusters": [
                JobCluster(
                    job_cluster_key="Remorph_Reconciliation_Cluster",
                    new_cluster=compute.ClusterSpec(
                        data_security_mode=compute.DataSecurityMode.USER_ISOLATION,
                        spark_conf={},
                        node_type_id=self._get_default_node_type_id(),
                        autoscale=compute.AutoScale(min_workers=2, max_workers=10),
                        spark_version=latest_lts_spark,
                    ),
                )
            ],
            "tasks": [
                self._job_recon_task(
                    Task(
                        task_key=task_key,
                        description=description,
                        job_cluster_key="Remorph_Reconciliation_Cluster",
                    ),
                    recon_config,
                    remorph_wheel_path,
                ),
            ],
            "max_concurrent_runs": 2,
            "parameters": [JobParameterDefinition(name="operation_name", default="reconcile")],
        }

    def _job_recon_task(self, jobs_task: Task, recon_config: ReconcileConfig, remorph_wheel_path: str) -> Task:
        libraries = [
            compute.Library(whl=remorph_wheel_path),
        ]

        if recon_config.data_source == ReconSourceType.ORACLE.value:
            # TODO: Automatically fetch a version list for `ojdbc8`
            oracle_driver_version = "23.4.0.24.05"
            libraries.append(
                compute.Library(
                    maven=compute.MavenLibrary(f"com.oracle.database.jdbc:ojdbc8:{oracle_driver_version}"),
                ),
            )

        return dataclasses.replace(
            jobs_task,
            libraries=libraries,
            python_wheel_task=PythonWheelTask(
                package_name="databricks_labs_remorph",
                entry_point="reconcile",
                parameters=["{{job.parameters.[operation_name]}}"],
            ),
        )

    def _is_testing(self):
        return self._product_info.product_name() != "remorph"

    @staticmethod
    def _get_test_purge_time() -> str:
        return (datetime.now(timezone.utc) + _TEST_JOBS_PURGE_TIMEOUT).strftime("%Y%m%d%H")

    def _get_default_node_type_id(self) -> str:
        return self._ws.clusters.select_node_type(local_disk=True, min_memory_gb=16)

    def _name_with_prefix(self, name: str) -> str:
        prefix = self._installation.product()
        return f"{prefix.upper()}_{name}".replace(" ", "_")



import asyncio
import dataclasses
import itertools
import json
import logging
import os
import re
import time
from collections.abc import Mapping
from pathlib import Path
from typing import NoReturn

from databricks.sdk.service.sql import CreateWarehouseRequestWarehouseType
from databricks.sdk import WorkspaceClient

from databricks.labs.blueprint.cli import App
from databricks.labs.blueprint.entrypoint import get_logger, is_in_debug
from databricks.labs.blueprint.installation import RootJsonValue
from databricks.labs.blueprint.tui import Prompts


from databricks.labs.lakebridge.assessments.configure_assessment import (
    create_assessment_configurator,
    PROFILER_SOURCE_SYSTEM,
)

from databricks.labs.lakebridge.config import TranspileConfig
from databricks.labs.lakebridge.contexts.application import ApplicationContext
from databricks.labs.lakebridge.helpers.recon_config_utils import ReconConfigPrompts
from databricks.labs.lakebridge.helpers.telemetry_utils import make_alphanum_or_semver
from databricks.labs.lakebridge.install import installer
from databricks.labs.lakebridge.reconcile.runner import ReconcileRunner
from databricks.labs.lakebridge.lineage import lineage_generator
from databricks.labs.lakebridge.reconcile.recon_config import RECONCILE_OPERATION_NAME, AGG_RECONCILE_OPERATION_NAME
from databricks.labs.lakebridge.transpiler.execute import transpile as do_transpile
from databricks.labs.lakebridge.transpiler.lsp.lsp_engine import LSPEngine
from databricks.labs.lakebridge.transpiler.repository import TranspilerRepository
from databricks.labs.lakebridge.transpiler.sqlglot.sqlglot_engine import SqlglotEngine
from databricks.labs.lakebridge.transpiler.transpile_engine import TranspileEngine

from databricks.labs.lakebridge.transpiler.transpile_status import ErrorSeverity

lakebridge = App(__file__)
logger = get_logger(__file__)


def raise_validation_exception(msg: str) -> NoReturn:
    raise ValueError(msg)


def _create_warehouse(ws: WorkspaceClient) -> str:

    dbsql = ws.warehouses.create_and_wait(
        name=f"lakebridge-warehouse-{time.time_ns()}",
        warehouse_type=CreateWarehouseRequestWarehouseType.PRO,
        cluster_size="Small",  # Adjust size as needed
        auto_stop_mins=30,  # Auto-stop after 30 minutes of inactivity
        enable_serverless_compute=True,
        max_num_clusters=1,
    )

    if dbsql.id is None:
        raise RuntimeError(f"Failed to create warehouse {dbsql.name}")

    logger.info(f"Created warehouse with id: {dbsql.id}")
    return dbsql.id


def _remove_warehouse(ws: WorkspaceClient, warehouse_id: str):
    ws.warehouses.delete(warehouse_id)
    logger.info(f"Removed warehouse post installation with id: {warehouse_id}")


@lakebridge.command
def transpile(
    *,
    w: WorkspaceClient,
    transpiler_config_path: str | None = None,
    source_dialect: str | None = None,
    input_source: str | None = None,
    output_folder: str | None = None,
    error_file_path: str | None = None,
    skip_validation: str | None = None,
    catalog_name: str | None = None,
    schema_name: str | None = None,
    transpiler_repository: TranspilerRepository = TranspilerRepository.user_home(),
):
    """Transpiles source dialect to databricks dialect"""
    ctx = ApplicationContext(w)
    logger.debug(f"Preconfigured transpiler config: {ctx.transpile_config!r}")
    ctx.add_user_agent_extra("cmd", "execute-transpile")
    checker = _TranspileConfigChecker(ctx.transpile_config, ctx.prompts, transpiler_repository)
    checker.use_transpiler_config_path(transpiler_config_path)
    checker.use_source_dialect(source_dialect)
    checker.use_input_source(input_source)
    checker.use_output_folder(output_folder)
    checker.use_error_file_path(error_file_path)
    checker.use_skip_validation(skip_validation)
    checker.use_catalog_name(catalog_name)
    checker.use_schema_name(schema_name)
    config, engine = checker.check()
    logger.debug(f"Final configuration for transpilation: {config!r}")

    assert config.source_dialect is not None, "Source dialect has been validated by this point."
    ctx.add_user_agent_extra("transpiler_source_tech", make_alphanum_or_semver(config.source_dialect))
    plugin_name = engine.transpiler_name
    plugin_name = re.sub(r"\s+", "_", plugin_name)
    ctx.add_user_agent_extra("transpiler_plugin_name", plugin_name)
    user = ctx.current_user
    logger.debug(f"User: {user}")

    result = asyncio.run(_transpile(ctx, config, engine))
    # DO NOT Modify this print statement, it is used by the CLI to display results in GO Table Template
    print(json.dumps(result))


class _TranspileConfigChecker:
    """Helper class for the 'transpile' command to check and consolidate the configuration."""

    #
    # Configuration parameters can come from 3 sources:
    #  - Command-line arguments (e.g., --input-source, --output-folder, etc.)
    #  - The configuration file, stored in the user's workspace home directory.
    #  - User prompts.
    #
    # The conventions are:
    #  - Command-line arguments take precedence over the configuration file.
    #  - Prompting is a last resort, only used when a required configuration value has not been provided and does not
    #    have a default value.
    #  - An invalid value results in a halt, with the error message indicating the source of the invalid value. We do
    #    NOT attempt to recover from invalid values by looking for another source:
    #     - Prompting unexpectedly will break scripting and automation.
    #     - Using an alternate value will need to confusion because the behaviour will not be what the user expects.
    #
    # This ensures that we distinguish between:
    #  - Invalid command-line arguments:
    #    Resolution: fix the command-line argument value.
    #  - Invalid prompt responses:
    #    Resolution: provide a valid response to the prompt.
    #  - Invalid configuration file values:
    #    Resolution: fix the configuration file value, or provide the command-line argument to override it.
    #
    # Implementation details:
    #  - For command-line arguments and prompted values, we:
    #     - Log the raw values (prior to validation) at DEBUG level, using the repr() rendering.
    #     - Validate the values immediately, with the error message on failure mentioning the source of the value.
    #     - Only update the configuration if the validation passes.
    #  - Prompting only occurs when a value is required, but not provided via the command-line argument or the
    #    configuration file.
    #  - In addition to the above, a final validation of everything is required: this ensures that values from the
    #    configuration file are validated, and if we have a failure we know that's the source because other sources
    #    were already checked.
    #  - The interplay between the source dialect and the transpiler config path is handled with care:
    #      - The source dialect, needs to be consistent with the engine that transpiler config path, refers to.
    #      - The source dialect can be used to infer the transpiler config path.
    #
    # TODO: Refactor this class to eliminate a lof of the boilerplate and handle this more elegantly.

    _config: TranspileConfig
    """The workspace configuration for transpiling, updated from command-line arguments."""
    _prompts: Prompts
    """Prompting system, for requesting configuration that hasn't been provided."""
    _source_dialect_override: str | None = None
    """The source dialect provided on the command-line, if any."""
    _transpiler_repository: TranspilerRepository
    """The repository where available transpilers are installed."""

    def __init__(
        self,
        config: TranspileConfig | None,
        prompts: Prompts,
        transpiler_repository: TranspilerRepository,
    ) -> None:
        if config is None:
            logger.warning(
                "No workspace transpile configuration, use 'install-transpile' to (re)install and configure; using defaults for now."
            )
            config = TranspileConfig()
        self._config = config
        self._prompts = prompts
        self._transpiler_repository = transpiler_repository
        self._source_dialect_override = None

    @staticmethod
    def _validate_transpiler_config_path(transpiler_config_path: str, msg: str) -> None:
        """Validate the transpiler config path: it must be a valid path that exists."""
        # Note: the content is not validated here, but during loading of the engine.
        if not Path(transpiler_config_path).exists():
            raise_validation_exception(msg)

    def use_transpiler_config_path(self, transpiler_config_path: str | None) -> None:
        if transpiler_config_path is not None:
            logger.debug(f"Setting transpiler_config_path to: {transpiler_config_path!r}")
            self._validate_transpiler_config_path(
                transpiler_config_path,
                f"Invalid path for '--transpiler-config-path', does not exist: {transpiler_config_path}",
            )
            self._config = dataclasses.replace(self._config, transpiler_config_path=transpiler_config_path)

    def use_source_dialect(self, source_dialect: str | None) -> None:
        if source_dialect is not None:
            # Defer validation: depends on the transpiler config path, we'll deal with this later.
            logger.debug(f"Pending source_dialect override: {source_dialect!r}")
            self._source_dialect_override = source_dialect

    @staticmethod
    def _validate_input_source(input_source: str, msg: str) -> None:
        """Validate the input source: it must be a path that exists."""
        if not Path(input_source).exists():
            raise_validation_exception(msg)

    def use_input_source(self, input_source: str | None) -> None:
        if input_source is not None:
            logger.debug(f"Setting input_source to: {input_source!r}")
            self._validate_input_source(
                input_source, f"Invalid path for '--input-source', does not exist: {input_source}"
            )
            self._config = dataclasses.replace(self._config, input_source=input_source)

    def _prompt_input_source(self) -> None:
        prompted_input_source = self._prompts.question("Enter input SQL path (directory/file)").strip()
        logger.debug(f"Setting input_source to: {prompted_input_source!r}")
        self._validate_input_source(
            prompted_input_source, f"Invalid input source, path does not exist: {prompted_input_source}"
        )
        self._config = dataclasses.replace(self._config, input_source=prompted_input_source)

    def _check_input_source(self) -> None:
        config_input_source = self._config.input_source
        if config_input_source is None:
            self._prompt_input_source()
        else:
            self._validate_input_source(
                config_input_source, f"Invalid input source path configured, does not exist: {config_input_source}"
            )

    @staticmethod
    def _validate_output_folder(output_folder: str, msg: str) -> None:
        """Validate the output folder: it doesn't have to exist, but its parent must."""
        if not Path(output_folder).parent.exists():
            raise_validation_exception(msg)

    def use_output_folder(self, output_folder: str | None) -> None:
        if output_folder is not None:
            logger.debug(f"Setting output_folder to: {output_folder!r}")
            self._validate_output_folder(
                output_folder, f"Invalid path for '--output-folder', parent does not exist for: {output_folder}"
            )
            self._config = dataclasses.replace(self._config, output_folder=output_folder)

    def _prompt_output_folder(self) -> None:
        prompted_output_folder = self._prompts.question("Enter output folder path (directory)").strip()
        logger.debug(f"Setting output_folder to: {prompted_output_folder!r}")
        self._validate_output_folder(
            prompted_output_folder, f"Invalid output folder path, parent does not exist for: {prompted_output_folder}"
        )
        self._config = dataclasses.replace(self._config, output_folder=prompted_output_folder)

    def _check_output_folder(self) -> None:
        config_output_folder = self._config.output_folder
        if config_output_folder is None:
            self._prompt_output_folder()
        else:
            self._validate_output_folder(
                config_output_folder,
                f"Invalid output folder configured, parent does not exist for: {config_output_folder}",
            )

    @staticmethod
    def _validate_error_file_path(error_file_path: str | None, msg: str) -> None:
        """Value the error file path: it doesn't have to exist, but its parent must."""
        if error_file_path is not None and not Path(error_file_path).parent.exists():
            raise_validation_exception(msg)

    def use_error_file_path(self, error_file_path: str | None) -> None:
        if error_file_path is not None:
            logger.debug(f"Setting error_file_path to: {error_file_path!r}")
            self._validate_error_file_path(
                error_file_path, f"Invalid path for '--error-file-path', parent does not exist: {error_file_path}"
            )
            self._config = dataclasses.replace(self._config, error_file_path=error_file_path)

    def _check_error_file_path(self) -> None:
        config_error_file_path = self._config.error_file_path
        self._validate_error_file_path(
            config_error_file_path,
            f"Invalid error file path configured, parent does not exist for: {config_error_file_path}",
        )

    def use_skip_validation(self, skip_validation: str | None) -> None:
        if skip_validation is not None:
            skip_validation_lower = skip_validation.lower()
            if skip_validation_lower not in {"true", "false"}:
                msg = f"Invalid value for '--skip-validation': {skip_validation!r} must be 'true' or 'false'."
                raise_validation_exception(msg)
            new_skip_validation = skip_validation_lower == "true"
            logger.debug(f"Setting skip_validation to: {new_skip_validation!r}")
            self._config = dataclasses.replace(self._config, skip_validation=new_skip_validation)

    def use_catalog_name(self, catalog_name: str | None) -> None:
        if catalog_name:
            logger.debug(f"Setting catalog_name to: {catalog_name!r}")
            self._config = dataclasses.replace(self._config, catalog_name=catalog_name)

    def use_schema_name(self, schema_name: str | None) -> None:
        if schema_name:
            logger.debug(f"Setting schema_name to: {schema_name!r}")
            self._config = dataclasses.replace(self._config, schema_name=schema_name)

    def _configure_transpiler_config_path(self, source_dialect: str) -> TranspileEngine | None:
        """Configure the transpiler config path based on the requested source dialect."""
        # Names of compatible transpiler engines for the given dialect.
        compatible_transpilers = self._transpiler_repository.transpilers_with_dialect(source_dialect)
        match len(compatible_transpilers):
            case 0:
                # Nothing found for the specified dialect, fail.
                return None
            case 1:
                # Only one transpiler available for the specified dialect, use it.
                transpiler_name = next(iter(compatible_transpilers))
                logger.debug(f"Using only transpiler available for dialect {source_dialect!r}: {transpiler_name!r}")
            case _:
                # Multiple transpilers available for the specified dialect, prompt for which to use.
                logger.debug(
                    f"Multiple transpilers available for dialect {source_dialect!r}: {compatible_transpilers!r}"
                )
                transpiler_name = self._prompts.choice("Select the transpiler:", list(compatible_transpilers))
        transpiler_config_path = self._transpiler_repository.transpiler_config_path(transpiler_name)
        logger.info(f"Lakebridge will use the {transpiler_name} transpiler.")
        self._config = dataclasses.replace(self._config, transpiler_config_path=str(transpiler_config_path))
        return LSPEngine.from_config_path(transpiler_config_path)

    def _configure_source_dialect(
        self, source_dialect: str, engine: TranspileEngine | None, msg_prefix: str
    ) -> TranspileEngine:
        """Configure the source dialect, if possible, and return the transpiler engine."""
        if engine is None:
            engine = self._configure_transpiler_config_path(source_dialect)
            if engine is None:
                supported_dialects = ", ".join(self._transpiler_repository.all_dialects())
                msg = f"{msg_prefix}: {source_dialect!r} (supported dialects: {supported_dialects})"
                raise_validation_exception(msg)
            else:
                self._config = dataclasses.replace(self._config, source_dialect=source_dialect)
        else:
            # Check the source dialect against the engine.
            if source_dialect not in engine.supported_dialects:
                supported_dialects_description = ", ".join(engine.supported_dialects)
                msg = f"Invalid value for '--source-dialect': {source_dialect!r} must be one of: {supported_dialects_description}"
                raise_validation_exception(msg)
            self._config = dataclasses.replace(self._config, source_dialect=source_dialect)
        return engine

    def _prompt_source_dialect(self) -> TranspileEngine:
        # This is similar to the post-install prompting for the source dialect.
        supported_dialects = self._transpiler_repository.all_dialects()
        match len(supported_dialects):
            case 0:
                msg = "No transpilers are available, install using 'install-transpile' or use --transpiler-conf-path'."
                raise_validation_exception(msg)
            case 1:
                # Only one dialect available, use it.
                source_dialect = next(iter(supported_dialects))
                logger.debug(f"Using only source dialect available: {source_dialect!r}")
            case _:
                # Multiple dialects available, prompt for which to use.
                logger.debug(f"Multiple source dialects available, choice required: {supported_dialects!r}")
                source_dialect = self._prompts.choice("Select the source dialect:", list(supported_dialects))
        engine = self._configure_transpiler_config_path(source_dialect)
        assert engine is not None, "No transpiler engine available for a supported dialect; configuration is invalid."
        self._config = dataclasses.replace(self._config, source_dialect=source_dialect)
        return engine

    def _check_lsp_engine(self) -> TranspileEngine:
        #
        # This is somewhat complicated:
        #  - If there is no transpiler config path, we need to try to infer it from the source dialect.
        #  - If there is no source dialect, we need to prompt for it: but that depends on the transpiler config path.
        #
        # With this in mind, the steps here are:
        # 1. If the transpiler config path is set, check it exists and load the engine.
        # 2. If the source dialect is set,
        #      - If the transpiler config path is set: validate the source dialect against the engine.
        #      - If the transpiler config path is not set: search for a transpiler that satisfies the dialect:
        #          * If one is found, we're good to go.
        #          * If more than one is found, prompt for the transpiler config path.
        #          * If none are found, fail: no transpilers available for the specified dialect.
        #    At this point we have either halted, or we have a valid transpiler path and source dialect.
        # 3. If the source dialect is not set, we need to:
        #      a) Load the set of available dialects: just for the engine if transpiler config path is set, or for all
        #         available transpilers if not.
        #      b) Depending on the available dialects:
        #          - If there is only one dialect available, set it as the source dialect.
        #          - If there are multiple dialects available, prompt for which to use.
        #          - If there are no dialects available, fail: no transpilers available.
        #    At this point we have either halted, or we have a valid transpiler path and source dialect.
        #
        # TODO: Deal with the transpiler options, and filtering them for the engine.
        #

        # Step 1: Check the transpiler config path.
        engine: TranspileEngine | None
        transpiler_config_path = self._config.transpiler_config_path
        if transpiler_config_path is not None:
            self._validate_transpiler_config_path(
                transpiler_config_path,
                f"Error: Invalid value for '--transpiler-config-path': '{str(transpiler_config_path)}', file does not exist.",
            )
            path = Path(transpiler_config_path)
            engine = LSPEngine.from_config_path(path)
        else:
            engine = None
        del transpiler_config_path

        # Step 2: Check the source dialect, assuming it has been specified, and infer the transpiler config path if necessary.
        source_dialect = self._source_dialect_override
        if source_dialect is not None:
            logger.debug(f"Setting source_dialect override: {source_dialect!r}")
            engine = self._configure_source_dialect(source_dialect, engine, "Invalid value for '--source-dialect'")
        else:
            source_dialect = self._config.source_dialect
            if source_dialect is not None:
                logger.debug(f"Using configured source_dialect: {source_dialect!r}")
                engine = self._configure_source_dialect(source_dialect, engine, "Invalid configured source dialect")
            else:
                # Step 3: Source dialect is not set, we need to prompt for it.
                logger.debug("No source_dialect available, prompting.")
                engine = self._prompt_source_dialect()
        return engine

    def _check_transpiler_options(self, engine: TranspileEngine) -> None:
        if not isinstance(engine, LSPEngine):
            return
        assert self._config.source_dialect is not None, "Source dialect must be set before checking transpiler options."
        options_for_dialect = engine.options_for_dialect(self._config.source_dialect)
        transpiler_options = self._config.transpiler_options
        if not isinstance(transpiler_options, Mapping):
            return
        checked_options = {
            option.flag: (
                transpiler_options[option.flag]
                if option.flag in transpiler_options
                else option.prompt_for_value(self._prompts)
            )
            for option in options_for_dialect
        }
        self._config = dataclasses.replace(self._config, transpiler_options=checked_options)

    def check(self) -> tuple[TranspileConfig, TranspileEngine]:
        """Checks that all configuration parameters are present and valid."""
        logger.debug(f"Checking config: {self._config!r}")

        self._check_input_source()
        self._check_output_folder()
        self._check_error_file_path()
        # No validation here required for:
        #   - skip_validation: it is a boolean flag, mandatory, and has a default: so no further validation is needed.
        #   - catalog_name and schema_name: they are mandatory, but have a default.
        # TODO: if validation is enabled, we should check that the catalog and schema names are valid.

        # This covers: transpiler_config_path, source_dialect
        engine = self._check_lsp_engine()

        # Last thing: the configuration may have transpiler-specific options, check them.
        self._check_transpiler_options(engine)

        config = self._config
        logger.debug(f"Validated config: {config!r}")
        return config, engine


async def _transpile(ctx: ApplicationContext, config: TranspileConfig, engine: TranspileEngine) -> RootJsonValue:
    """Transpiles source dialect to databricks dialect"""
    ctx.add_user_agent_extra("cmd", "execute-transpile")
    user = ctx.current_user
    logger.debug(f"User: {user}")
    _override_workspace_client_config(ctx, config.sdk_config)
    status, errors = await do_transpile(ctx.workspace_client, engine, config)

    logger.debug(f"Transpilation completed with status: {status}")

    for path, errors_by_path in itertools.groupby(errors, key=lambda x: x.path):
        errs = list(errors_by_path)
        errors_by_severity = {
            severity.name: len(list(errors)) for severity, errors in itertools.groupby(errs, key=lambda x: x.severity)
        }
        reports = []
        reported_severities = [ErrorSeverity.ERROR, ErrorSeverity.WARNING]
        for severity in reported_severities:
            if severity.name in errors_by_severity:
                word = str.lower(severity.name) + "s" if errors_by_severity[severity.name] > 1 else ""
                reports.append(f"{errors_by_severity[severity.name]} {word}")

        msg = ", ".join(reports) + " found"

        if ErrorSeverity.ERROR.name in errors_by_severity:
            logger.error(f"{path}: {msg}")
        elif ErrorSeverity.WARNING.name in errors_by_severity:
            logger.warning(f"{path}: {msg}")

    # Table Template in labs.yml requires the status to be list of dicts Do not change this
    return [status]


def _override_workspace_client_config(ctx: ApplicationContext, overrides: dict[str, str] | None) -> None:
    """
    Override the Workspace client's SDK config with the user provided SDK config.
    Users can provide the cluster_id and warehouse_id during the installation.
    This will update the default config object in-place.
    """
    if not overrides:
        return

    warehouse_id = overrides.get("warehouse_id")
    if warehouse_id:
        ctx.connect_config.warehouse_id = warehouse_id

    cluster_id = overrides.get("cluster_id")
    if cluster_id:
        ctx.connect_config.cluster_id = cluster_id


@lakebridge.command
def reconcile(*, w: WorkspaceClient) -> None:
    """[EXPERIMENTAL] Reconciles source to Databricks datasets"""
    ctx = ApplicationContext(w)
    ctx.add_user_agent_extra("cmd", "execute-reconcile")
    user = ctx.current_user
    logger.debug(f"User: {user}")
    recon_runner = ReconcileRunner(
        ctx.workspace_client,
        ctx.installation,
        ctx.install_state,
        ctx.prompts,
    )
    recon_runner.run(operation_name=RECONCILE_OPERATION_NAME)


@lakebridge.command
def aggregates_reconcile(*, w: WorkspaceClient) -> None:
    """[EXPERIMENTAL] Reconciles Aggregated source to Databricks datasets"""
    ctx = ApplicationContext(w)
    ctx.add_user_agent_extra("cmd", "execute-aggregates-reconcile")
    user = ctx.current_user
    logger.debug(f"User: {user}")
    recon_runner = ReconcileRunner(
        ctx.workspace_client,
        ctx.installation,
        ctx.install_state,
        ctx.prompts,
    )

    recon_runner.run(operation_name=AGG_RECONCILE_OPERATION_NAME)


@lakebridge.command
def generate_lineage(
    *,
    w: WorkspaceClient,
    source_dialect: str | None = None,
    input_source: str,
    output_folder: str,
) -> None:
    """[Experimental] Generates a lineage of source SQL files or folder"""
    ctx = ApplicationContext(w)
    logger.debug(f"User: {ctx.current_user}")
    if not os.path.exists(input_source):
        raise_validation_exception(f"Invalid path for '--input-source': Path '{input_source}' does not exist.")
    if not os.path.exists(output_folder):
        raise_validation_exception(f"Invalid path for '--output-folder': Path '{output_folder}' does not exist.")
    if source_dialect is None:
        raise_validation_exception("Value for '--source-dialect' must be provided.")
    engine = SqlglotEngine()
    supported_dialects = engine.supported_dialects
    if source_dialect not in supported_dialects:
        supported_dialects_description = ", ".join(supported_dialects)
        msg = f"Unsupported source dialect provided for '--source-dialect': '{source_dialect}' (supported: {supported_dialects_description})"
        raise_validation_exception(msg)

    lineage_generator(engine, source_dialect, input_source, output_folder)


@lakebridge.command
def configure_secrets(*, w: WorkspaceClient) -> None:
    """Setup reconciliation connection profile details as Secrets on Databricks Workspace"""
    recon_conf = ReconConfigPrompts(w)

    # Prompt for source
    source = recon_conf.prompt_source()

    logger.info(f"Setting up Scope, Secrets for `{source}` reconciliation")
    recon_conf.prompt_and_save_connection_details()


@lakebridge.command(is_unauthenticated=True)
def configure_database_profiler() -> None:
    """[Experimental] Install the lakebridge Assessment package"""
    prompts = Prompts()

    # Prompt for source system
    source_system = str(
        prompts.choice("Please select the source system you want to configure", PROFILER_SOURCE_SYSTEM)
    ).lower()

    # Create appropriate assessment configurator
    assessment = create_assessment_configurator(source_system=source_system, product_name="lakebridge", prompts=prompts)
    assessment.run()


@lakebridge.command
def install_transpile(
    *,
    w: WorkspaceClient,
    artifact: str | None = None,
    transpiler_repository: TranspilerRepository = TranspilerRepository.user_home(),
) -> None:
    """Install or upgrade the Lakebridge transpilers."""
    ctx = ApplicationContext(w)
    ctx.add_user_agent_extra("cmd", "install-transpile")
    if artifact:
        ctx.add_user_agent_extra("artifact-overload", Path(artifact).name)
    user = w.current_user
    logger.debug(f"User: {user}")
    transpile_installer = installer(w, transpiler_repository)
    transpile_installer.run(module="transpile", artifact=artifact)


@lakebridge.command(is_unauthenticated=False)
def configure_reconcile(
    *,
    w: WorkspaceClient,
    transpiler_repository: TranspilerRepository = TranspilerRepository.user_home(),
) -> None:
    """Configure the Lakebridge reconciliation module"""
    ctx = ApplicationContext(w)
    ctx.add_user_agent_extra("cmd", "configure-reconcile")
    user = w.current_user
    logger.debug(f"User: {user}")
    if not w.config.warehouse_id:
        dbsql_id = _create_warehouse(w)
        w.config.warehouse_id = dbsql_id
    logger.debug(f"Warehouse ID used for configuring reconcile: {w.config.warehouse_id}.")
    reconcile_installer = installer(w, transpiler_repository)
    reconcile_installer.run(module="reconcile")


@lakebridge.command
def analyze(
    *,
    w: WorkspaceClient,
    source_directory: str | None = None,
    report_file: str | None = None,
    source_tech: str | None = None,
):
    """Run the Analyzer"""
    ctx = ApplicationContext(w)
    ctx.add_user_agent_extra("cmd", "analyze")

    logger.debug(f"User: {ctx.current_user}")
    ctx.analyzer.run_analyzer(source_directory, report_file, source_tech)


if __name__ == "__main__":
    lakebridge()
    if is_in_debug():
        logger.setLevel(logging.DEBUG)



import logging
import os
import sys
import subprocess

from databricks.labs.blueprint.entrypoint import find_project_root
from databricks.labs.blueprint.cli import App


def proxy_command(app: App, command: str):
    def fn(**_):
        proxy = JvmProxy()
        proxy.run()

    fn.__name__ = command
    fn.__doc__ = f"Proxy to run {command} in JVM"
    app.command(fn)


class JvmProxy:
    # TODO Refactor this class to use LSP protocol instead
    def __init__(self):
        self._root = find_project_root(__file__)
        databricks_logger = logging.getLogger("databricks")
        self._debug = databricks_logger.level == logging.DEBUG

    def _recompile(self):
        subprocess.run(
            ["mvn", "compile", "-f", f'{self._root}/pom.xml'],
            stdout=sys.stdout,
            stderr=sys.stderr,
            check=True,
        )

    def run(self):
        if self._debug:
            self._recompile()
        classpath = self._root / 'core/target/classpath.txt'
        classes = self._root / 'core/target/scala-2.12/classes'
        # TODO: use the os-specific path separator
        args = [
            "java",
            "--class-path",
            f'{classes.as_posix()}:{classpath.read_text()}',
            "com.databricks.labs.lakebridge.Main",
            sys.argv[1],
        ]
        with subprocess.Popen(
            args,
            stdin=sys.stdin,
            stdout=sys.stdout,
            stderr=sys.stderr,
            env=os.environ.copy(),
            text=True,
        ) as process:
            return process.wait()



from pathlib import Path

from databricks.labs.blueprint.wheels import ProductInfo
from databricks.labs.lakebridge.coverage import commons
from databricks.labs.lakebridge.transpiler.sqlglot.generator.databricks import Databricks
from databricks.labs.lakebridge.transpiler.sqlglot.parsers.snowflake import Snowflake

if __name__ == "__main__":
    input_dir = commons.get_env_var("INPUT_DIR_PARENT", required=True)
    output_dir = commons.get_env_var("OUTPUT_DIR", required=True)

    REMORPH_COMMIT_HASH = commons.get_current_commit_hash() or ""  # C0103 pylint
    product_info = ProductInfo(__file__)
    remorph_version = product_info.unreleased_version()

    if not input_dir:
        raise ValueError("Environment variable `INPUT_DIR_PARENT` is required")
    if not output_dir:
        raise ValueError("Environment variable `OUTPUT_DIR` is required")

    commons.collect_transpilation_stats(
        "Remorph",
        REMORPH_COMMIT_HASH,
        remorph_version,
        Snowflake,
        Databricks,
        Path(input_dir) / 'snowflake',
        Path(output_dir),
    )



import logging

from functools import reduce
import sqlglot.expressions as exp
from sqlglot import Dialect

from databricks.labs.lakebridge.reconcile.query_builder.base import QueryBuilder
from databricks.labs.lakebridge.reconcile.query_builder.expression_generator import (
    build_column,
    concat,
    get_hash_transform,
    lower,
    transform_expression,
    build_column_no_alias,
)

logger = logging.getLogger(__name__)


def _hash_transform(
    node: exp.Expression,
    source: Dialect,
    layer: str,
):
    transform = get_hash_transform(source, layer)
    return transform_expression(node, transform)


_HASH_COLUMN_NAME = "hash_value_recon"


class HashQueryBuilder(QueryBuilder):

    def build_query(self, report_type: str) -> str:

        if report_type != 'row':
            self._validate(self.join_columns, f"Join Columns are compulsory for {report_type} type")

        _join_columns = self.join_columns if self.join_columns else set()
        hash_cols = sorted((_join_columns | self.select_columns) - self.threshold_columns - self.drop_columns)

        key_cols = hash_cols if report_type == "row" else sorted(_join_columns | self.partition_column)

        cols_with_alias = [self._build_column_with_alias(col) for col in key_cols]

        # in case if we have column mapping, we need to sort the target columns in the order of source columns to get
        # same hash value
        hash_cols_with_alias = [
            {"this": self._build_column_name_source_normalized(col), "alias": self._build_alias_source_normalized(col)}
            for col in hash_cols
        ]
        sorted_hash_cols_with_alias = sorted(hash_cols_with_alias, key=lambda column: column["alias"])
        hashcols_sorted_as_src_seq = [column["this"] for column in sorted_hash_cols_with_alias]

        key_cols_with_transform = (
            self._apply_user_transformation(cols_with_alias) if self.user_transformations else cols_with_alias
        )
        hash_col_with_transform = [self._generate_hash_algorithm(hashcols_sorted_as_src_seq, _HASH_COLUMN_NAME)]

        res = (
            exp.select(*hash_col_with_transform + key_cols_with_transform)
            .from_(":tbl")
            .where(self.filter, dialect=self.engine)
            .sql(dialect=self.engine)
        )

        logger.info(f"Hash Query for {self.layer}: {res}")
        return res

    def _generate_hash_algorithm(
        self,
        cols: list[str],
        column_alias: str,
    ) -> exp.Expression:
        cols_no_alias = [build_column_no_alias(this=col) for col in cols]
        cols_with_transform = self.add_transformations(cols_no_alias, self.engine)
        col_exprs = exp.select(*cols_with_transform).iter_expressions()
        concat_expr = concat(list(col_exprs))

        if self.engine == "oracle":
            concat_expr = reduce(lambda x, y: exp.DPipe(this=x, expression=y), concat_expr.expressions)

        hash_expr = concat_expr.transform(_hash_transform, self.engine, self.layer).transform(lower, is_expr=True)

        return build_column(hash_expr, alias=column_alias)



from databricks.sdk.useragent import alphanum_pattern, semver_pattern


def make_alphanum_or_semver(value: str) -> str:
    if alphanum_pattern.match(value) or semver_pattern.match(value):
        return value
    # assume it's not a semver, replace illegal alphanum chars
    result = []
    for char in value:
        if not alphanum_pattern.match(char):
            char = '_'
        result.append(char)
    return "".join(result)
